from .predict import predict 
import argparse
import sys, multiprocessing
import torch

def _parse_args():
    parser=argparse.ArgumentParser(description="Run SolTranNet aqueous solubility predictor")
    parser.add_argument('input',nargs='?',type=argparse.FileType('r'),default=sys.stdin,help='PATH to the file containing the SMILES you wish to use. Assumes the content is 1 SMILE per line.')
    parser.add_argument('output',nargs='?',type=argparse.FileType('w'),default=sys.stdout,help='Name of the output file. Defaults to stdout.')
    parser.add_argument('--batchsize',default=32,type=int,help='Batch size for the data loader. Defaults to 32.')
    parser.add_argument('--cpus',default=multiprocessing.cpu_count(),type=int,help='Number of CPU cores to use for the data loader. Defaults to use all available cores. Pass 0 to only run on 1 CPU.')
    parser.add_argument('--cpu_predict',action='store_true',help='Flag to force the predictions to be made on only the CPU. Default behavior is to use GPU if available.')

    args=parser.parse_args()

    return args

def _run(args):

    smiles=[x.rstrip() for x in args.input]
    if args.cpu_predict:
        predictions=predict(smiles,batch_size=args.batchsize,num_workers=args.cpus,device=torch.device('cpu'))
    else:
        predictions=predict(smiles,batch_size=args.batchsize,num_workers=args.cpus)
    for pred, smi, warn in predictions:
        args.output.write(f'{smi},{pred:.3f},{warn}\n')




"""
Code based on:
Maziarka et al "Molecule Attention Transformer" -> https://github.com/ardigen/MAT
"""

import math, copy
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


### Model definition -- the default values are the architecture for SolTranNet
def make_model(d_atom=28, N=8, d_model=8, h=2, dropout=0.1, lambda_attention=0.5,
               N_dense=1, leaky_relu_slope=0.0, aggregation_type='mean', 
               dense_output_nonlinearity='relu',n_output=1,
               scale_norm=False, init_type='uniform', use_adapter=False, n_generator_layers=1):
    "Helper: Construct a model from hyperparameters."
    c = copy.deepcopy
    attn = MultiHeadedAttention(h, d_model, dropout, lambda_attention)
    ff = PositionwiseFeedForward(d_model, N_dense, dropout, leaky_relu_slope, dense_output_nonlinearity)
    model = GraphTransformer(
        Encoder(EncoderLayer(d_model, c(attn), c(ff), dropout, scale_norm, use_adapter), N, scale_norm),
        Embeddings(d_model, d_atom, dropout),
        Generator(d_model, aggregation_type, n_output, n_generator_layers, leaky_relu_slope, dropout, scale_norm))

    return model


class GraphTransformer(nn.Module):
    def __init__(self, encoder, src_embed, generator):
        super(GraphTransformer, self).__init__()
        self.encoder = encoder
        self.src_embed = src_embed
        self.generator = generator
        
    def forward(self, src, src_mask, adj_matrix, edges_att):
        "Take in and process masked src and target sequences."
        return self.predict(self.encode(src, src_mask, adj_matrix, edges_att), src_mask)
    
    def encode(self, src, src_mask, adj_matrix, edges_att):
        return self.encoder(self.src_embed(src), src_mask, adj_matrix, edges_att)
    
    def predict(self, out, out_mask):
        return self.generator(out, out_mask)
    
    
class Generator(nn.Module):
    "Define standard linear + softmax generation step."
    def __init__(self, d_model, aggregation_type='mean', n_output=1, n_layers=1, 
                 leaky_relu_slope=0.01, dropout=0.0, scale_norm=False):
        super(Generator, self).__init__()
        if n_layers == 1:
            self.proj = nn.Linear(d_model, n_output)
        else:
            self.proj = []
            for i in range(n_layers-1):
                self.proj.append(nn.Linear(d_model, d_model))
                self.proj.append(nn.LeakyReLU(leaky_relu_slope))
                self.proj.append(ScaleNorm(d_model) if scale_norm else LayerNorm(d_model))
                self.proj.append(nn.Dropout(dropout))
            self.proj.append(nn.Linear(d_model, n_output))
            self.proj = torch.nn.Sequential(*self.proj)
        self.aggregation_type = aggregation_type

    def forward(self, x, mask):
        mask = mask.unsqueeze(-1).float()
        out_masked = x * mask
        if self.aggregation_type == 'mean':
            out_sum = out_masked.sum(dim=1)
            mask_sum = mask.sum(dim=(1))
            out_avg_pooling = out_sum / mask_sum
        elif self.aggregation_type == 'sum':
            out_sum = out_masked.sum(dim=1)
            out_avg_pooling = out_sum
        elif self.aggregation_type == 'dummy_node':
            out_avg_pooling = out_masked[:,0]
        projected = self.proj(out_avg_pooling)
        return projected
    
    
class PositionGenerator(nn.Module):
    "Define standard linear + softmax generation step."
    def __init__(self, d_model):
        super(PositionGenerator, self).__init__()
        self.norm = LayerNorm(d_model)
        self.proj = nn.Linear(d_model, 3)

    def forward(self, x, mask):
        mask = mask.unsqueeze(-1).float()
        out_masked = self.norm(x) * mask
        projected = self.proj(out_masked)
        return projected
    

### Encoder

def clones(module, N):
    "Produce N identical layers."
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class Encoder(nn.Module):
    "Core encoder is a stack of N layers"
    def __init__(self, layer, N, scale_norm):
        super(Encoder, self).__init__()
        self.layers = clones(layer, N)
        self.norm = ScaleNorm(layer.size) if scale_norm else LayerNorm(layer.size)
        
    def forward(self, x, mask, adj_matrix, edges_att):
        "Pass the input (and mask) through each layer in turn."
        for layer in self.layers:
            x = layer(x, mask, adj_matrix, edges_att)
        return self.norm(x)

    
class LayerNorm(nn.Module):
    "Construct a layernorm module (See citation for details)."
    def __init__(self, features, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.a_2 = nn.Parameter(torch.ones(features))
        self.b_2 = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2
    
    
class ScaleNorm(nn.Module):
    """ScaleNorm"""
    "All g’s in SCALE NORM are initialized to sqrt(d)"
    def __init__(self, scale, eps=1e-5):
        super(ScaleNorm, self).__init__()
        self.scale = nn.Parameter(torch.tensor(math.sqrt(scale)))
        self.eps = eps
        
    def forward(self, x):
        norm = self.scale / torch.norm(x, dim=-1, keepdim=True).clamp(min=self.eps)
        return x * norm

    
class SublayerConnection(nn.Module):
    """
    A residual connection followed by a layer norm.
    Note for code simplicity the norm is first as opposed to last.
    """
    def __init__(self, size, dropout, scale_norm, use_adapter):
        super(SublayerConnection, self).__init__()
        self.norm = ScaleNorm(size) if scale_norm else LayerNorm(size)
        self.dropout = nn.Dropout(dropout)
        self.use_adapter = use_adapter
        self.adapter = Adapter(size, 8) if use_adapter else None

    def forward(self, x, sublayer):
        "Apply residual connection to any sublayer with the same size."
        if self.use_adapter:
            return x + self.dropout(self.adapter(sublayer(self.norm(x))))
        return x + self.dropout(sublayer(self.norm(x)))

    
class EncoderLayer(nn.Module):
    "Encoder is made up of self-attn and feed forward (defined below)"
    def __init__(self, size, self_attn, feed_forward, dropout, scale_norm, use_adapter):
        super(EncoderLayer, self).__init__()
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(size, dropout, scale_norm, use_adapter), 2)
        self.size = size

    def forward(self, x, mask, adj_matrix, edges_att):
        "Follow Figure 1 (left) for connections."
        x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, adj_matrix, edges_att, mask))
        return self.sublayer[1](x, self.feed_forward)

    
### Attention
def attention(query, key, value, adj_matrix, edges_att,
              mask=None, dropout=None, lambdas=(0.5, 0.5),
              eps=1e-6, inf=1e12):
    "Compute 'Scaled Dot Product Attention'"
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) \
             / math.sqrt(d_k)
    if mask is not None:
        scores = scores.masked_fill(mask.unsqueeze(1).repeat(1, query.shape[1], query.shape[2], 1) == 0, -inf)
    p_attn = F.softmax(scores, dim = -1)

    # Prepare adjacency matrix
    adj_matrix = adj_matrix / (adj_matrix.sum(dim=-1).unsqueeze(2) + eps)
    adj_matrix = adj_matrix.unsqueeze(1).repeat(1, query.shape[1], 1, 1)
    p_adj = adj_matrix
    
    lambda_attention, lambda_adjacency = lambdas
    p_weighted = lambda_attention * p_attn + lambda_adjacency * p_adj
    
    if dropout is not None:
        p_weighted = dropout(p_weighted)

    atoms_features = torch.matmul(p_weighted, value)     
    return atoms_features, p_weighted, p_attn


class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1, lambda_attention=0.5):
        "Take in model size and number of heads."
        super(MultiHeadedAttention, self).__init__()
        assert d_model % h == 0
        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        lambda_adjacency = 1. - lambda_attention
        self.lambdas = (lambda_attention, lambda_adjacency)
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)
        
    def forward(self, query, key, value, adj_matrix, edges_att, mask=None):
        "Implements Figure 2"
        if mask is not None:
            # Same mask applied to all h heads.
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)
        
        # 1) Do all the linear projections in batch from d_model => h x d_k 
        query, key, value = \
            [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for l, x in zip(self.linears, (query, key, value))]
        
        # 2) Apply attention on all the projected vectors in batch. 
        x, self.attn, self.self_attn = attention(query, key, value, adj_matrix, edges_att,
                                                 mask=mask, dropout=self.dropout, lambdas=self.lambdas)
        
        # 3) "Concat" using a view and apply a final linear. 
        x = x.transpose(1, 2).contiguous() \
             .view(nbatches, -1, self.h * self.d_k)
        return self.linears[-1](x)


### Conv 1x1 aka Positionwise feed forward

class PositionwiseFeedForward(nn.Module):
    "Implements FFN equation."
    def __init__(self, d_model, N_dense, dropout=0.1, leaky_relu_slope=0.0, dense_output_nonlinearity='relu'):
        super(PositionwiseFeedForward, self).__init__()
        self.N_dense = N_dense
        self.linears = clones(nn.Linear(d_model, d_model), N_dense)
        self.dropout = clones(nn.Dropout(dropout), N_dense)
        self.leaky_relu_slope = leaky_relu_slope
        if dense_output_nonlinearity == 'relu':
            self.dense_output_nonlinearity = lambda x: F.leaky_relu(x, negative_slope=self.leaky_relu_slope)
        elif dense_output_nonlinearity == 'tanh':
            self.tanh = torch.nn.Tanh()
            self.dense_output_nonlinearity = lambda x: self.tanh(x)
        elif dense_output_nonlinearity == 'none':
            self.dense_output_nonlinearity = lambda x: x
            

    def forward(self, x):
        if self.N_dense == 0:
            return x
        
        for i in range(len(self.linears)-1):
            x = self.dropout[i](F.leaky_relu(self.linears[i](x), negative_slope=self.leaky_relu_slope))
            
        return self.dropout[-1](self.dense_output_nonlinearity(self.linears[-1](x)))

    
## Embeddings

class Embeddings(nn.Module):
    def __init__(self, d_model, d_atom, dropout):
        super(Embeddings, self).__init__()
        self.lut = nn.Linear(d_atom, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.lut(x))



from unittest import TestCase
from soltrannet.data_utils import load_data_from_smiles
from soltrannet.predict import check_smile

class TestWarning(TestCase):
    def test_working_warnings(self):
        smiles=["c1ccccc1","Cn1cnc2n(C)c(=O)n(C)c(=O)c12","c1ccccc1 .ignore","[Zn+2]","[Na+].[Cl-]"]
        features=load_data_from_smiles(smiles)
        node_features=[x[0] for x in features]
        warnings=[]
        for smi, f in zip(smiles,node_features):
            warnings.append(check_smile(smi,f))

        self.assertTrue(warnings==['','','','Other-typed Atom(s) Detected Prediction less reliable','Salt Other-typed Atom(s) Detected Prediction less reliable'])



from unittest import TestCase
import soltrannet as stn
import torch
import numpy as np

class TestPredictions(TestCase):
    def test_preds(self):

        def helper(a,b):
            a_array=np.array([x[0] for x in a])
            b_array=np.array([x[0] for x in b])

            first = np.any(np.isclose(a_array,b_array))
            second = [x[1] for x in a] == [x[1] for x in b]
            third = [x[2] for x in a] == [x[2] for x in b]

            return first and second and third

        smiles=["c1ccccc1","[Zn+2]","[Na+].[Cl-]"]
        correct=[(-1.052748441696167, 'c1ccccc1', ''),(-6.881845474243164,'[Zn+2]','Other-typed Atom(s) Detected Prediction less reliable'),(-0.16869020462036133,'[Na+].[Cl-]','Salt Other-typed Atom(s) Detected Prediction less reliable')]
        
        test_cpu=list(stn.predict(smiles,device='cpu'))
        self.assertTrue(helper(correct, test_cpu))

        if torch.cuda.is_available():
            test_gpu=list(stn.predict(smiles,device='cuda'))
            self.assertTrue(helper(correct, test_gpu))



"""
Code based on:
Shang et al "Edge Attention-based Multi-Relational Graph Convolutional Networks" -> https://github.com/Luckick/EAGCN
Coley et al "Convolutional Embedding of Attributed Molecular Graphs for Physical Property Prediction" -> https://github.com/connorcoley/conv_qsar_fast
"""

import sys
try:
    from rdkit import Chem
    from rdkit.Chem import MolFromSmiles
except:
    sys.exit('rdkit is not installed. Install with:\nconda install -c rdkit rdkit')

import logging
import numpy as np
import torch
import sys
from torch.utils.data import Dataset

use_cuda = torch.cuda.is_available()
FloatTensor = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor
LongTensor = torch.cuda.LongTensor if use_cuda else torch.LongTensor
IntTensor = torch.cuda.IntTensor if use_cuda else torch.IntTensor
DoubleTensor = torch.cuda.DoubleTensor if use_cuda else torch.DoubleTensor


def load_data_from_smiles(x_smiles, add_dummy_node=True):
    """Load and featurize data from lists of SMILES strings and labels.

    Args:
        x_smiles (list[str]): A list of SMILES strings.
        add_dummy_node (bool): If True, a dummy node will be added to the molecular graph. Defaults to True.
        one_hot_formal_charge (bool): If True, formal charges on atoms are one-hot encoded. Defaults to True.

    Returns:
        A tuple of lists of graph descriptors (node features, adjacency matrices)
    """

    x_all = []
    for smiles in x_smiles:
        try:
            mol = MolFromSmiles(smiles)
            afm, adj = featurize_mol(mol, add_dummy_node)
            x_all.append([afm, adj])
        except ValueError as e:
            logging.warning('the SMILES ({}) can not be converted to a graph.\nREASON: {}'.format(smiles,e))

    return x_all

def featurize_mol(mol, add_dummy_node):
    """Featurize molecule.

    Args:
        mol (rdchem.Mol): An RDKit Mol object.
        add_dummy_node (bool): If True, a dummy node will be added to the molecular graph.

    Returns:
        A tuple of molecular graph descriptors (node features, adjacency matrix).
    """
    node_features = np.array([get_atom_features(atom)
                              for atom in mol.GetAtoms()])

    adj_matrix = np.eye(mol.GetNumAtoms())
    for bond in mol.GetBonds():
        begin_atom = bond.GetBeginAtom().GetIdx()
        end_atom = bond.GetEndAtom().GetIdx()
        adj_matrix[begin_atom, end_atom] = adj_matrix[end_atom, begin_atom] = 1

    if add_dummy_node:
        m = np.zeros((node_features.shape[0] + 1, node_features.shape[1] + 1))
        m[1:, 1:] = node_features
        m[0, 0] = 1.
        node_features = m

        m = np.zeros((adj_matrix.shape[0] + 1, adj_matrix.shape[1] + 1))
        m[1:, 1:] = adj_matrix
        adj_matrix = m

    return node_features, adj_matrix


#lookup table for one-hot elements
anummap = {5:0, 6: 1, 7:2, 8: 3, 9: 4,  15:5, 16:6, 17:7, 35:8, 53:9 }
anumtable = np.full(128,10)
for i,val in anummap.items():
    anumtable[i] = val
    
def get_atom_features(atom):
    """Calculate atom features.

            Identity            -- [B,C,N,O,F,P,S,Cl,Br,I,Dummy,Other]
            #Heavy Neighbors    -- [0,1,2,3,4,5]
            #H atoms            -- [0,1,2,3,4]
            Formal Charge       -- [-1,0,1]
            Is in a Ring        -- [0,1]
            Is Aromatic         -- [0,1]
        Dummy and Other types, have the same one-hot encoding, but the dummy node is unconnected.
    Args:
        atom (rdchem.Atom): An RDKit Atom object.

    Returns:
        A 1-dimensional array (ndarray) of atom features.
    """
    attributes = np.zeros(27)    
    anum = atom.GetAtomicNum()        
    attributes[anumtable[anum]] = 1.0
        
    ncnt = min(len(atom.GetNeighbors()),5)
    attributes[11+ncnt] = 1.0

    hcnt = min(atom.GetTotalNumHs(),4)
    attributes[17+hcnt] = 1.0
    
    charge = atom.GetFormalCharge()
    if charge == 0:
        attributes[23] = 1.0
    elif charge < 0:
        attributes[22] = 1.0
    else:
        attributes[24] = 1.0
        
    attributes[25] = atom.IsInRing()
    attributes[26] = atom.GetIsAromatic()
    return attributes


def one_hot_vector(val, lst):
    """Converts a value to a one-hot vector based on options in lst"""
    if val not in lst:
        val = lst[-1]
    return map(lambda x: x == val, lst)


class Molecule:
    """
        Class that represents a train/validation/test datum
        - self.label: 0 neg, 1 pos -1 missing for different target.
    """

    def __init__(self, x, index, smile=''):
        self.node_features = x[0]
        self.adjacency_matrix = x[1]
        self.index = index
        self.smile = smile


class MolDataset(Dataset):
    """
    Class that represents a train/validation/test dataset that's readable for PyTorch
    Note that this class inherits torch.utils.data.Dataset
    """

    def __init__(self, data_list):
        """
        @param data_list: list of Molecule objects
        """
        self.data_list = data_list

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, key):
        if type(key) == slice:
            return MolDataset(self.data_list[key])
        return self.data_list[key]


def pad_array(array, shape, dtype=np.float32):
    """Pad a 2-dimensional array with zeros.

    Args:
        array (ndarray): A 2-dimensional array to be padded.
        shape (tuple[int]): The desired shape of the padded array.
        dtype (data-type): The desired data-type for the array.

    Returns:
        A 2-dimensional array of the given shape padded with zeros.
    """
    padded_array = np.zeros(shape, dtype=dtype)
    padded_array[:array.shape[0], :array.shape[1]] = array
    return padded_array


def mol_collate_func(batch):
    """Create a padded batch of molecule features.

    Args:
        batch (list[Molecule]): A batch of raw molecules.

    Returns:
        A list of FloatTensors with padded molecule features: (adjacency matrices, node features).
    """
    adjacency_list, features_list, smiles_list, index_list = [], [], [], []

    max_size = 0
    for molecule in batch:
        if molecule.adjacency_matrix.shape[0] > max_size:
            max_size = molecule.adjacency_matrix.shape[0]

    for molecule in batch:
        adjacency_list.append(pad_array(molecule.adjacency_matrix, (max_size, max_size)))
        features_list.append(pad_array(molecule.node_features, (max_size, molecule.node_features.shape[1])))
        smiles_list.append(molecule.smile)
        index_list.append(molecule.index)

    #do not use cuda memory during data loading
    return [torch.FloatTensor(adjacency_list), torch.FloatTensor(features_list), smiles_list, index_list]

def construct_dataset(x_all):
    """Construct a MolDataset object from the provided data.

    Args:
        x_all (list): A list of molecule features.

    Returns:
        A MolDataset object filled with the provided data.
    """
    output = [Molecule(data, i)
              for i, data in enumerate(x_all)]
    return MolDataset(output)

def construct_loader(x, batch_size=32, shuffle=False):
    """Construct a data loader for the provided data.

    Args:
        x (list): A list of molecule features.
        batch_size (int): The batch size. Defaults to 32
        shuffle (bool): If True the data will be loaded in a random order. Defaults to False.

    Returns:
        A DataLoader object that yields batches of padded molecule features.
    """
    data_set = construct_dataset(x)
    loader = torch.utils.data.DataLoader(dataset=data_set,batch_size=batch_size,collate_fn=mol_collate_func,shuffle=shuffle)
    return loader


class MolIterableDataset(torch.utils.data.IterableDataset):
    '''An iterable dataset over Molecule objects.'''
    
    def __init__(self, x_smiles, add_dummy_node=True):
        '''Initialize iterable dataset with list of smiles'''
        super(MolIterableDataset).__init__()
        self.x_smiles = x_smiles
        self.add_dummy_node=add_dummy_node
                
    def __iter__(self):
        
        worker_info = torch.utils.data.get_worker_info()
        if worker_info:
            index = worker_info.id
            nworkers = worker_info.num_workers
        else:
            index = 0
            nworkers = 1
                    
        for i,smiles in enumerate(self.x_smiles):
            try:
                if i%nworkers == index:
                    mol = MolFromSmiles(smiles.split()[0])
                    afm, adj = featurize_mol(mol, self.add_dummy_node)
                    yield Molecule([afm, adj],i,smiles)
            except ValueError as e:
                logging.warning('the SMILES ({}) can not be converted to a graph.\nREASON: {}'.format(smiles,e))


def construct_loader_from_smiles(smiles, batch_size=32, num_workers=1, shuffle=False):
    """Construct a data loader for the provided data.

    Args:
        smiles (list): A list of smiles.
        batch_size (int): The batch size. Defaults to 32
        shuffle (bool): If True the data will be loaded in a random order. Defaults to False.

    Returns:
        A DataLoader object that yields batches of padded molecule features.
    """
    data_set = MolIterableDataset(smiles)
    loader = torch.utils.data.DataLoader(dataset=data_set,batch_size=batch_size,collate_fn=mol_collate_func,shuffle=shuffle,num_workers=num_workers)
    return loader



import argparse
from unittest import TestCase
from soltrannet import _run
import io

class TestCommandLine(TestCase):
    def test_command_line(self):
        with io.StringIO() as buf:
            correct='c1ccccc1,-1.053,\nc1ccccc1 .ignore,-1.053,\nCn1cnc2n(C)c(=O)n(C)c(=O)c12,-1.132,\n[Zn+2],-6.882,Other-typed Atom(s) Detected Prediction less reliable\n[Na+].[Cl-],-0.169,Salt Other-typed Atom(s) Detected Prediction less reliable\n'

            args=argparse.Namespace(input=['c1ccccc1','c1ccccc1 .ignore','Cn1cnc2n(C)c(=O)n(C)c(=O)c12','[Zn+2]','[Na+].[Cl-]'],
                                                                                             output=buf,batchsize=5, cpus=0,cpu_predict=None)
            _run(args)
            self.assertEqual(buf.getvalue(),correct)



import torch
import numpy as np
import pkg_resources, heapq

from .data_utils import construct_loader_from_smiles
from .transformer import make_model

#initialize model on import
_weights=pkg_resources.resource_filename(__name__,"soltrannet_aqsol_trained.weights")
_model=make_model()

if torch.cuda.is_available():
    _device=torch.device("cuda")
    _model.load_state_dict(torch.load(_weights))    
    _model.to(_device)
else:
    _device=torch.device('cpu')
    _model.load_state_dict(torch.load(_weights,map_location=_device))

    
def predict(smi, batch_size=32, num_workers=1,device=_device):
    """Predict Solubilities for a list of SMILES.
    Args:
        smi ([str]): A list of SMILES strings, upon which we wish to predict the solubilities for.
        batch_size: sizes of batches to use
        num_workers: number of parallel workers to use 
        device: torch device to use
    Returns:
        A list of tuples [(prediction, SMILES, Warning)].
    """
    #generate the molecular graphs from the SMILES strings
    data_loader = construct_loader_from_smiles(smi, batch_size=batch_size, num_workers=num_workers)
    
    #Then we ensure the model is set up properly
    _model.to(device)
    _model.eval()

    #Now we can generate our predictions.
    #Molecules are processed out of order due to parallelism, so use heap
    #to yield molecules back in order
    H = []
    index = 0
    with torch.no_grad():
        for batch in data_loader:
            adjacency_matrix, node_features, smiles, indices = batch
            adjacency_matrix = adjacency_matrix.to(device)
            node_features = node_features.to(device)
            
            batch_mask = torch.sum(torch.abs(node_features),dim=-1) != 0
            pred = _model(node_features, batch_mask, adjacency_matrix, None)
            for pred,node_feature,smi,i in zip(pred.flatten().tolist(), node_features, smiles, indices):
                heapq.heappush(H,(i,(pred, smi, check_smile(smi, node_feature))))
                while H and H[0][0] == index:
                    index += 1
                    yield heapq.heappop(H)[1]
                    
    while H:
        yield heapq.heappop(H)[1]


def check_smile(smi, node_features):
    """Check the smile to see if the predictions will be less reliable. E.G. are a salt or have Other typed atoms

    Args:
        smile (str): A Smile string
        features [(node features, adjacency matrix)]: A list of the graph represenation for the corresponding SMILES string.

    Returns:
        A string with any warnings

    """
    warn=""

    #start with a check for salts
    if '.' in smi.split()[0]:
        warn+='Salt '

    #use the calculated molecular features to determine if an "Other" typed atom exists in the molecule.
    if node_features[:,11].sum():
        warn+='Other-typed Atom(s) '

    if warn!='':
        warn+='Detected Prediction less reliable'

    return warn



from setuptools import setup, find_packages
import pathlib

here=pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding='utf-8')

setup(

    name='soltrannet',

    version='1.0.0',

    description='Molecule attention transformer for predicting aqueous solubility (log S)',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/gnina/SolTranNet',

    author='Paul Francoeur',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='ML, solubility, tool',

    packages=find_packages(),

    python_requires='>=3.6, <4',

    install_requires=[
        'torch >= 1.7.0',
        'pathlib >= 1.0',
    ],

    package_data={
        'soltrannet':['soltrannet_aqsol_trained.weights'],
    },

    entry_points={
        'console_scripts':['soltrannet=soltrannet.command_line:main'],
    },

    test_suite='nose.collector',
    tests_require=['nose','nose-cover3'],

    project_urls={
        'Bug Reports': 'https://github.com/gnina/SolTranNet/issues',
        'Source': 'https://github.com/gnina/SolTranNet',
    },
)
